
#### Function: multiple_gbm ####
multiple_gbm <- function(formula, data_to_train, n_outputs,best_hyper,train.fraction) {
  # randomize data
  random_index <- sample(1:nrow(data_to_train), nrow(data_to_train))
  random_data_to_train <- data_to_train[random_index, ]
  list_gbm <- list()
  for (index in 1:n_outputs) {
    col_names_output <- paste0("PC",index)
    Vx_formula <- paste0( col_names_output,formula)
    gbm1 <- gbm(as.formula(Vx_formula),
                data = random_data_to_train,
                distribution = "gaussian",
                
                n.trees               = best_hyper$n.trees[index],          
                interaction.depth     = best_hyper$interaction.depth[index], 
                shrinkage             = best_hyper$shrinkage[index],  
                n.minobsinnode        = best_hyper$n.minobsinnode[index],    
                bag.fraction          = best_hyper$bag.fraction[index],    
                cv.folds              = best_hyper$cv.folds[index],                 
                keep.data             = best_hyper$keep.data[index], 
                train.fraction = train.fraction,  
                verbose = FALSE,
                n.cores = 1)
    
    list_gbm[[index]] <- gbm1
  }
  return(list_gbm)
}


#### Function: optimezed_multiple_gbm ####
# Goal : build surrogate model based on gradient boosting machine method for several outputs
# Created on: 15/12/2021
optimized_multiple_gbm <- function(formula,
                                   data_to_train,
                                   n_outputs,
                                   n.trees,
                                   hyper_grid,
                                   train.fraction,
                                   n.cores, # will use all cores by default
                                   verbose) {
  # randomize data
  random_index <- sample(1:nrow(data_to_train), nrow(data_to_train))
  random_data_to_train <- data_to_train[random_index, ] 
  list_gbm <- list()
  for (index in 1:n_outputs) {
    col_names_output <- paste0("PC",index)
    Vx_formula <- paste0( col_names_output,formula)
    gbm1 <- gbm(as.formula(Vx_formula),
                data = random_data_to_train,
                distribution = "gaussian",
                n.trees = n.trees,
                interaction.depth = hyper_grid$interaction.depth,
                shrinkage = hyper_grid$shrinkage,
                n.minobsinnode = hyper_grid$n.minobsinnode,
                bag.fraction = hyper_grid$bag.fraction,
                cv.folds = hyper_grid$cv.folds,                
                train.fraction = train.fraction,  
                keep.data = hyper_grid$keep.data, 
                verbose = verbose,
                n.cores = n.cores)  
    list_gbm[[index]] <- gbm1
  }
  return(list_gbm)
}




#### Function: multiple_predict ####
# Goal : predict an array of outputs from a list of surrogate models
# Created on: 9/11/2021
multiple_predict <- function(list_predict, newdata, n_outputs) {
  # list_predict : list of surrogate models used to create ouput array
  # newdata : array or data.frame containing model parameter values to evaluate the surrogate
  # n_outputs :  dimension of the output array 
  res <- NULL
  for (index in 1:n_outputs) {
    single_res <- predict( list_predict[[index]] , newdata=newdata )
    res <- cbind(res,single_res)
  }
  res
}

#### Function: surrogate_model ####
# Goal : Evaluate multidimensional surrogate model output 
# Created on: 9/11/2021
surrogate_model <- function(newdata, c_gbm, data_to_train.pca, n_outputs,i=NULL,log_data=NULL) {
  
  # newdata: array or data.frame containing model parameter values to evaluate the surrogate
  # c_gbm: multidimensional surogate model created by multiple_gbm (or similar functions: to be implmented)
  # data_to_train.pca: PCA cretaed using the training data
  # n_outputs :  dimension of the reduced output array
  # i: if provided provides only results for specific output dimension
  
  res <- multiple_predict( c_gbm, newdata=newdata , n_outputs)
  res <- res[,] %*% t(data_to_train.pca$rotation[,1:n_outputs])
  if(all(data_to_train.pca$scale != FALSE)){
    res  <- scale(res, center = FALSE , scale=1/data_to_train.pca$scale)
  }
  if(all(data_to_train.pca$center != FALSE)){
    res <- scale(res, center = -1 * data_to_train.pca$center, scale=FALSE)
  }
  
  
  if(!is.null(i)) {res <- res[,i]}
  #res <- newdata[,1]
  if(log_data) res <- exp(res)
  
  return(res)
  
}

#### Function: surrogate_model_onPC ####
# Goal : Evaluate multidimensional surrogate model output on PCA components 
# Created on: 17/11/2021
surrogate_model_onPC <- function(newdata,c_gbm,i=NULL) {
  res <- predict( c_gbm[[i]] , newdata=newdata )
  return(res)
}
#### Function: mulitple_sobol ####
# Goal : Compute sobol indices for each output dimension - Not yet optimized optimized:
# Created on: 9/11/2021
mulitple_sobol_onPC  <- function(surrogate_model_onPC,X1,X2,c_gbm)  
{
  S1 <- NULL
  ST <- NULL
  for (i in 1:n_outputs) {
    sobol_indexs <- sobol2002(model = surrogate_model_onPC,
                              X1 = X1,
                              X2 = X2,
                              c_gbm = c_gbm,
                              i=i)
    
    S1_local <- as.data.frame(t(as.matrix(sobol_indexs$S)))
    rownames(S1_local) <- i
    ST_local <- as.data.frame(t(as.matrix(sobol_indexs$T)))
    rownames(ST_local) <- i
    
    S1 <- rbind(S1,S1_local)
    ST <- rbind(ST,ST_local)
  }
  return(list(S1,ST))
}

#### Function: surrogate_model ####
# Goal : Evaluate multidimensional surrogate model output 
# Created on: 17/11/2021
surrogate_model_onPC2 <- function(newdata, c_gbm) {
  res <- multiple_predict( c_gbm, newdata=newdata , n_outputs)
  return(res)
}

mulitple_sobol_vjan <- function(surrogate_model,X1,X2,dim_output,...)  
{
  num_sample <- nrow(X1)
  num_param  <- ncol(X1)
  if ( (num_sample != nrow(X2)) | (num_param != ncol(X2)) ) print("OJO! X1 and X2 have different dimensions")

  f <- surrogate_model(X1,...)
  f2 <- f*f
  ff <- surrogate_model(X2,...)
  fff <- f*ff
  
  F0 <- sqrt( colSums(fff, na.rm = TRUE)/num_sample )
  D <- abs( colSums(f2, na.rm = TRUE)/num_sample - (F0*F0) )
  
  rm(fff,f2)
  
  Di <- numeric(length = dim_output)
#  dim(Di) <- c(dim_output)
  Dii <- Di
  ST <- numeric(length = dim_output)
  S1 <- ST
  
  S1_m <- numeric(length = dim_output*num_param)
  dim(S1_m) <- c(num_param,dim_output)
  ST_m <- S1_m
  
  # Variance decomposition
  for (i_param in 1:num_param) 
  {
    kesi1 <- X1
    kesi2 <- X2
    
    kesi3 <- kesi2
    kesi3[,i_param] <- kesi1[,i_param]
    
    kesi4=kesi1
    kesi4[,i_param] <- kesi2[,i_param]
    
    surr_3 <- surrogate_model(kesi3,...)
    Di  <- colSums( f * (surr_3-ff), na.rm = TRUE) / num_sample
    rm(surr_3)
    
    surr_4 <- surrogate_model(kesi4,...)
    Dii <- colSums( f * (surr_4-ff), na.rm = TRUE) / num_sample
    rm(surr_4)
    
    #S1[i_param,]  <- Di/D
    #ST[i_param,]   <- -Dii/D + 1
    
    S1 <- Di/D
    ST <- -Dii/D + 1
    saveRDS(S1, file = paste0("S1_",i_param,".rds"))
    saveRDS(ST, file = paste0("ST_",i_param,".rds"))
    
    S1_m[i_param,] <- S1
    ST_m[i_param,] <- ST
  }
  return(list(S1_m,ST_m))
}
#### Function: mulitple_sobol ####
# Goal : Compute sobol indices for each output dimension - Not yet optimized optimized:
# Created on: 9/11/2021
mulitple_sobol  <- function(surrogate_model,X1,X2,c_gbm,data_to_train.pca,n_dim,log_data)  
{
  S1 <- NULL
  ST <- NULL
  for (i in 1:n_dim) {
    
    sobol_indexs <- sobol2002(model = surrogate_mode,
                              X1 = X1,
                              X2 = X2,
                              c_gbm = c_gbm,
                              data_to_train.pca = data_to_train.pca,
                              n_outputs = n_outputs,
                              i=i,
                              log_data=log_data)
    
    S1_local <- as.data.frame(t(as.matrix(sobol_indexs$S)))
    rownames(S1_local) <- i
    ST_local <- as.data.frame(t(as.matrix(sobol_indexs$T)))
    rownames(ST_local) <- i
    
    S1 <- rbind(S1,S1_local)
    ST <- rbind(ST,ST_local)
  }
  return(list(S1,ST))
}

#### Function: plot_field ####
# Goal : Reshape data to 2D matrix for 1D Pisces fields OJO dimensions are hard-coded!!!
# Created on: 16/11/2021
plot_field <- function(var,i_row,data_for_plot )
{
  my.padding <- list(layout.heights = list(
    top.padding = 0,
    main.key.padding = 0,
    key.axis.padding = 0,
    axis.xlab.padding = 0,
    xlab.key.padding = 0,
    key.sub.padding = 0),
    layout.widths = list(
      left.padding = 0,
      key.ylab.padding = 0,
      ylab.axis.padding = 0,
      axis.key.padding = 0,
      right.padding = 0)
  )
  

  n_dim <- 47*73*8
  data_array <- data.matrix(data_for_plot[i_row,1:n_dim])
  dim(data_array) <- c(47,73,8)
  par(mar=c(0,0,0,0))
  matrix_to_plot <- t( data_array[dim(data_array)[1]:1,,var] )
  levelplot( matrix_to_plot,col.regions=viridis(2000), xlab =NULL, ylab = NULL, par.settings = my.padding)
  
}

plot_allfields <- function(S1_param){
  S1_X <-readRDS(paste0(S1_param,".rds"))
  print(paste(S1_param,"Mean:", mean(S1_X), "Max:",max(S1_X) ))
  dim(S1_X) <- c(1,length(S1_X))
  
  p1 <-  plot_field(1,1,S1_X )
  p2 <-  plot_field(2,1,S1_X )
  p3 <-  plot_field(3,1,S1_X )
  p4 <-  plot_field(4,1,S1_X )
  p5 <-  plot_field(5,1,S1_X )
  p6 <-  plot_field(6,1,S1_X )
  p7 <-  plot_field(7,1,S1_X )
  p8 <-  plot_field(8,1,S1_X )
  pdf(file = paste0(S1_param,".pdf"), width = 30, height = 2)
  grid.arrange(p1,p2,p3,p4,p5,p6,p7,p8, ncol=11)
  dev.off()
}
